# -*- coding: utf-8 -*-
"""Command-line interface functions for rpi2caster"""

# IMPORTS for text user interface
import io
import os
import readline
import glob
from . import exceptions as e
from . import constants

# Whether the debug mode is on (can be changed by setting module's attribute)
DEBUG_MODE = False
# Style modifiers for displaying bold, italic, smallcaps, inferior, superior
STYLE_MODIFIERS = {'roman': ' ',
                   'bold': '*',
                   'italic': '/',
                   'smallcaps': '#',
                   'subscript': '_',
                   'superscript': '^'}
# Some standard prompts
MSG_MENU = '[Enter] to go back to main menu...'
MSG_CONTINUE = '[Enter] to continue...'


def menu(options, header='', footer='', no_debug=False):
    """menu(options=[(name1, long1, func1), (name2, long2, func2)...],
                    header=foo,
                    footer=bar,
                    no_debug=False):

    A menu which takes four arguments:
    options list - contains tuples for each option:
        (description, long_description, function)
    header - string to be displayed above,
    footer - string to be displayed below,
    no_debug - True if we don't want to tell the user
        that the program is in debugging mode.

    After choice is made, return the command.

    Set up vars for conditional statements,
    and lists for appending new items.

    choices - options to be entered by user
    """
    # Clear the screen, display header and add two empty lines
    clear()
    if header:
        print(header, end='\n\n')
    # Get the first option - this will be listed last
    try:
        (zero_function, zero_desc, zero_long_desc) = options[0]
        functions = [zero_function]
    except IndexError:
        raise e.ExitProgram
    # Display all the options
    # Tab indent, option number, option name (not processing option 0 yet!)
    for i, (function, desc, long_desc) in enumerate(options):
        if i > 0:
            functions.append(function)
            print('\t %i : %s \n\t\t %s \n' % (i, desc, long_desc))
    # Option 0 is displayed last, add some whitespace around it
    print('\n\t %i : %s \n\t\t %s \n' % (0, zero_desc, zero_long_desc))
    # Print footer, if defined
    if footer:
        print(footer, end='\n\n')
    if DEBUG_MODE and not no_debug:
        print('The program is now in debugging mode!', end='\n\n')
    # Add an empty line to separate prompt
    print('\n')
    # Ask for user input
    choice_number = -1
    # Get only available options and exclude non-numeric strings
    while choice_number not in range(len(options)):
        # Wait until user enters proper data
        your_choice = input('Your choice: ')
        try:
            choice_number = int(your_choice)
        except ValueError:
            # Entered anything non-digit - repeat
            your_choice = ''
    # At last, we have chosen a valid option...
    # Return a corresponding value - which is option
    return functions[choice_number]


def clear():
    """Clears the screen"""
    os.system('clear')


def display(*args, **kwargs):
    """Displays info for the user - print all in one line"""
    print(*args, **kwargs)


def display_header(text, symbol='-'):
    """Displays a header banner"""
    dash_line = symbol * len(text)
    print('\n\n' + dash_line + '\n' + text + '\n' + dash_line + '\n')


def display_parameters(*data):
    """Displays the parameters given as a list/tuple of 2-item lists/tuples:
    [(value1, description1), (value2, description2)] - if value is there"""
    print('\n'.join(['%s: %s' % (desc, value) for item in data
                     for (value, desc) in item if value]) + '\n')


def debug_info(*args, **kwargs):
    """Prints debug messages to screen if in debug mode"""
    if DEBUG_MODE:
        print('DEBUG: ', *args, **kwargs)


def debug_enter_data(prompt):
    """For debug-specific data"""
    if DEBUG_MODE:
        return input('DEBUG: ' + prompt)


def debug_pause(msg1='', msg2=MSG_CONTINUE):
    """For debug confirmations"""
    if DEBUG_MODE:
        input('DEBUG: ' + msg1 + ' - ' + msg2)


def pause(msg1='', msg2=MSG_CONTINUE):
    """Waits until user presses return"""
    input(msg1 + '\n' + msg2)


def enter_data(prompt, datatype=str):
    """Enter a value and convert it to the specific datatype"""
    value = ''
    while not value:
        value = input(prompt)
        try:
            value = datatype(value)
        except ValueError:
            print('Incorrect value or data type!')
            value = ''
    return value


def enter_data_or_blank(prompt, datatype=str):
    """enter_data_or_blank:

    Enter a value or leave blank, try to convert to the specified datatype
    """
    while True:
        value = input(prompt)
        if not value:
            return value
        try:
            return datatype(value)
        except ValueError:
            print('Incorrect value or data type!')


def _format_display(character, style):
    """format_display:

    Uses ANSI escape sequences to alter the appearance of the character
    displayed in the matrix case layout.

    Temporarily unused - until bugfix for 24574.
    """
    style_codes = {'roman': 0,
                   'bold': 1,
                   'italic': 33,
                   'smallcaps': 34,
                   'subscript': 35,
                   'superscript': 36}
    closing_sequence = '\033[0m'
    try:
        starting_sequence = '\033[' + str(style_codes[style]) + 'm'
    except KeyError:
        starting_sequence = closing_sequence
    character = starting_sequence + character + closing_sequence
    print(character)
    return character


def format_display(character, style):
    """This is a placeholder to be used until Python bug 24574 is fixed"""
    try:
        return STYLE_MODIFIERS[style] + character
    except KeyError:
        return character


def tab_complete(text, state):
    """tab_complete(text, state):

    This function enables tab key auto-completion when you
    enter the filename.
    """
    return (glob.glob(text+'*')+[None])[state]


def enter_input_filename():
    """Allows to enter the input filename and checks if it is readable"""
    # Set readline parameters
    readline.set_completer_delims(' \t\n;')
    readline.parse_and_bind('tab: complete')
    readline.set_completer(tab_complete)
    # Enter the input filename; check if the file is readable
    while True:
        prompt = '\nEnter the input file name (leave blank to abort): '
        filename = enter_data_or_blank(prompt)
        if not filename:
            return False
        filename = os.path.realpath(filename)
        try:
            with io.open(filename, 'r'):
                return filename
        except (IOError, FileNotFoundError):
            print('Wrong filename or file not readable!\n')


def enter_output_filename():
    """Allows user to enter output filename (without checking if readable)"""
    # Set readline parameters
    readline.set_completer_delims(' \t\n;')
    readline.parse_and_bind('tab: complete')
    readline.set_completer(tab_complete)
    # Enter the output filename; no check here
    prompt = '\nEnter the input file name (leave blank to abort): '
    filename = enter_data_or_blank(prompt)
    if filename:
        return os.path.realpath(filename)
    else:
        return False


def simple_menu(message, options):
    """Simple menu:

    A simple menu where user is asked what to do.
    Wrong choice points back to the menu.

    Message: prompt - string displayed on screen;
    options: a dict {ans1:opt1, ans2:opt2...}.
    """
    ans = ''
    while True:
        ans = input(message)
        if ans in options:
            return options[ans]
        elif ans.lower() in options:
            return options[ans.lower()]
        elif ans.upper() in options:
            return options[ans.upper()]
        else:
            pass


def confirm(question):
    """Asks a simple question with yes or no answers.
    Returns True for yes and False for no."""
    return simple_menu('%s [Y / N]: ' % question, {'Y': True, 'N': False})


def display_diecase_layout(diecase):
    """display_diecase_layout:

    Shows a layout for a given diecase ID.
    Allows to specify a stopbar/wedge unit arrangement for this diecase,
    or uses the typical S5 if not specified.
    """
    # Define subroutines
    def process_matrix(mat):
        """Modifies matrix for displaying"""
        # Mat is defined as (char, (style1, style2...), column, row, units)
        (character, styles, column, row, units) = mat
        # Display different spaces as symbols
        spaces_symbols = {'_': '▣', ' ': '□', '': ' '}
        # Low space
        if character in spaces_symbols:
            character = spaces_symbols[character]
        # Otherwise we have a character - modify how it's displayed
        # based on style(s)
        else:
            try:
                for style in styles:
                    character = format_display(character, style)
            except (IndexError, KeyError):
                pass
        # Add column and row to sets
        cols_set.add(column)
        rows_set.add(row)
        # Finish
        return (character, styles, column, row, units)
    # Initialize columns and rows sets
    cols_set = rows_set = set()
    # Build a list of all characters
    all_mats = [process_matrix(mat) for mat in diecase.layout]
    # Build rows and columns to iterate over
    col_numbers = ((16 in rows_set or 'NI' in cols_set or 'NL' in cols_set) and
                   constants.COLUMNS_17 or constants.COLUMNS_15)
    # If row 16 found - generate 16 rows; else 15
    row_numbers = [x for x in range(1, 16 in rows_set and 17 or 16)]
    # Arrange matrices for displaying
    # Generate a header with column numbers
    header = ('|Row|' + ''.join([col.center(4) for col in col_numbers]) +
              '|Units|Shift|')
    # "-----" line in the table
    separator = '—' * len(header)
    # A row with only spaces and vertical lines in it
    empty_row = ('|' + ' ' * 3 + '|' +
                 ' ' * 4 * len(col_numbers) + '|' +
                 ' ' * 5 + '|' + ' ' * 5 + '|')
    # Initialize the displayed layout
    displayed_layout = [separator, header, separator, empty_row]
    # Process each row
    for row_number in row_numbers:
        # Get unit width value of the wedge for this row
        units = diecase.wedge.unit_arrangement[row_number] or ''
        shifted_units = diecase.wedge.unit_arrangement[row_number-1] or ''
        # Start with row number...
        row = ['|' + str(row_number).center(3) + '|']
        # Add only characters and styles, center chars to 4
        row.extend([mat[0].center(4)
                    for column_number in col_numbers for mat in all_mats
                    if mat[2] == column_number and mat[3] == row_number])
        row.append('|' + str(units).center(5) + '|')
        row.append(str(shifted_units).center(5) + '|')
        displayed_layout.append(''.join(row))
        displayed_layout.append(empty_row)
    # Add the header at the bottom
    displayed_layout.extend([separator, header, separator])
    # We can display it now
    for row in displayed_layout:
        print(row)
    # Explanation of symbols
    print('\nExplanation:', '□ = low space, ▣ = high space',
          '*a = bold, /a = italic, #a = small caps',
          '_a = subscript (inferior), ^a = superscript (superior)',
          sep='\n', end='\n\n')


def edit_diecase_layout(diecase):
    """edit_diecase_layout(diecase):

    Edits a matrix case layout, row by row, matrix by matrix. Allows to enter
    a position to be edited.
    """
    # Deep-copy a diecase to create an independent object to work on
    working_diecase = diecase.copy()

    def get_matrix(column, row):
        """Gets matrix data for given coordinates."""
        mat = [m for m in working_diecase.layout
               if column == m[2] and row == m[3]][0]
        return mat

    def display_matrix_details(mat):
        """Displays details for a given mat"""
        (char, styles, column, row, units) = mat
        print('\nDetails for matrix at %s%s:' % (column, row))
        print('Character: %s' % char)
        print('Styles: %s' % ', '.join([style for style in styles]))
        print('Unit width: %s' % units)
        print()

    def change_parameters(mat):
        """Edits a single mat in the diecase. Returns matrix description."""
        (char, styles, column, row, units) = mat
        # Edit it?
        print('Enter character: " " for low space (typical), "_" for '
              'high space (less common), leave empty to exit...')
        char = (enter_data_or_blank('Character?: ', str) or
                e.menu_level_up())
        available_styles = {'r': 'roman', 'b': 'bold',
                            'i': 'italic', 's': 'smallcaps',
                            'l': 'subscript', 'u': 'superscript'}
        print('Assign all text styles this matrix will be used for. '
              'More than one style is OK - e.g. roman and small caps.\n'
              'Available styles: [r]oman, [b]old, [i]talic, [s]mall caps,\n'
              '[l]ower index (a.k.a. subscript, inferior), '
              '[u]pper index (a.k.a. superscript, superior).\n'
              'Leave blank for roman only.')
        styles = enter_data_or_blank('Styles?: ', str) or 'r'
        styles = [available_styles[char] for char in styles
                  if char in available_styles]
        print('How many units for this character? '
              'Leave blank for normal wedge step value')
        units = enter_data_or_blank('Units?: ', int) or 0
        # Matrix is defined, return the data
        return [char, styles, column, row, units]

    def save_matrix(mat):
        """Updates the diecase layout with the new matrix data"""
        (_, _, column, row, _) = mat
        # Get current matrix data
        old_mat = get_matrix(column, row)
        mat_id = working_diecase.layout.index(old_mat)
        if confirm('Save the matrix in layout?'):
            working_diecase.layout[mat_id] = mat

    def edit_matrix(mat):
        """Displays a matrix info, asks for confirmation, edits the mat"""
        prompt = ('Edit this matrix: [Y]es / [N]o / '
                  '[F]inish editing? ')
        options = {'Y': True, 'N': False, 'F': 'exit'}
        # Display, ask, edit, save - or do nothing
        display_matrix_details(mat)
        decision = simple_menu(prompt, options)
        if decision == 'exit':
            e.menu_level_up()
        elif decision:
            # Edit the mat
            try:
                mat = change_parameters(mat)
                save_matrix(mat)
            except e.MenuLevelUp:
                pass

    def single_cell_mode():
        """Allows to specify a cell by its coordinates and edit it."""
        col_prompt = 'Column [NI, NL, A...O] or [Enter] to exit]? :'
        while True:
            try:
                column = ''
                row = 0
                while column not in constants.COLUMNS_17:
                    column = enter_data_or_blank(col_prompt, str)
                    column = column.upper() or e.menu_level_up()
                while row not in range(1, 17):
                    row = enter_data('Row (1 - 16)?: ', int)
                mat = get_matrix(column, row)
                edit_matrix(mat)
            except e.MenuLevelUp:
                break

    def all_rows_mode():
        """Row-by-row editing - all cells in row 1, then 2 etc."""
        try:
            for mat in working_diecase.layout:
                edit_matrix(mat)
        except e.MenuLevelUp:
            pass

    def all_columns_mode():
        """Column-by-column editing - all cells in column NI, NL, A...O"""
        # Rearrange the layout so that it's read column by column
        try:
            (_,) = [edit_matrix(mat) for col in constants.COLUMNS_17
                    for mat in working_diecase.layout if mat[2] == col]
        except e.MenuLevelUp:
            pass

    def single_row_mode():
        """Edits matrices found in a single row"""
        while True:
            try:
                row = 0
                while row not in range(1, 17):
                    row = enter_data_or_blank('Row (1 - 16)?: ', int)
                    row = row or e.menu_level_up()
                (_,) = [edit_matrix(mat) for mat in working_diecase.layout
                        if mat[3] == row]
            except e.MenuLevelUp:
                break

    def single_column_mode():
        """Edits matrices found in a single column"""
        col_prompt = 'Column [NI, NL, A...O] or [Enter] to exit]? :'
        while True:
            try:
                column = ''
                while column not in constants.COLUMNS_17:
                    column = enter_data_or_blank(col_prompt, str)
                    column = column.upper() or e.menu_level_up()
                (_,) = [edit_matrix(mat) for mat in working_diecase.layout
                        if mat[2] == column]
            except e.MenuLevelUp:
                break
    # Map unit values to rows
    # If the layout is empty, we need to initialize it
    print('\nCurrent diecase layout:\n')
    display_diecase_layout(working_diecase)
    prompt = ('\nChoose edit mode or press [Enter] to quit:\n'
              'AR - all matrices row by row,\n'
              'AC - all matrices column by column,\n'
              'M - single matrix by coordinates,\n'
              'R - all matrices in a specified row,\n'
              'C - all matrices in a specified column.'
              '\nYour choice:')
    options = {'AR': all_rows_mode,
               'AC': all_columns_mode,
               'M': single_cell_mode,
               'R': single_row_mode,
               'C': single_column_mode,
               '': e.menu_level_up}
    while True:
        try:
            simple_menu(prompt, options)()
        except e.MenuLevelUp:
            que = 'Save the changes?'
            if working_diecase.layout != diecase.layout and confirm(que):
                diecase.layout = working_diecase.layout
            break


def exit_program(*_):
    """Exit program:

    All objects call this method whenever they want to exit program.
    This is because we may do something specific in different UIs,
    so an abstraction layer may come in handy.
    """
    print('\n\nGoodbye!\n')
    exit()



# -*- coding: utf-8 -*-
"""
wedge_data

Contains higher-level functions for accessing wedge data.
This sits on top of the database module and is an abstraction layer
for other modules (like inventory, casting, typesetter).
Processes the data retrieved from database.
"""
from copy import deepcopy
# Default user interface
from .global_settings import UI
# Custom exceptions for rpi2caster suite
from . import exceptions as e
# Constants for known normal wedge unit arrangements
from . import wedge_arrangements
# Database backend
from . import database
DB = database.Database()


class Wedge(object):
    """Default S5-12E wedge"""
    def __init__(self):
        # Default wedge data - for S5-12E
        self.series = '5'
        self.set_width = 12
        self.is_brit_pica = True
        self.unit_arrangement = wedge_arrangements.S5

    @property
    def pica(self):
        """Get the pica value for the wedge. Can be .1667" (old British)
        or .166" (new British; American). The .1667" was commonly used with
        wedges made for European markets (wedge designation series-setE).
        Curiously, the old British pica is the same as modern DTP pica."""
        return self.is_brit_pica and 0.1667 or 0.166

    @property
    def parameters(self):
        """Gets a list of parameters"""
        return [('\n', '\nWedge data'),
                (self.series, 'Wedge series'),
                (self.set_width, 'Set width'),
                (self.is_brit_pica, 'British pica (.1667") based wedge?'),
                (' '.join([str(x) for x in self.unit_arrangement if x]),
                 'Unit arrangement for this wedge')]

    def copy(self):
        """Copies itself and returns an independent object"""
        return deepcopy(self)

    def edit(self, wedge_name=None):
        """Defines a wedge based on designation"""
        # Ask for wedge name and set width as it is written on the wedge
        prompt = ('Wedge name or leave blank to abort: ')
        # For countries that use comma as decimal delimiter, convert to point:
        wedge_name = wedge_name or UI.enter_data_or_blank(prompt)
        wedge_name = wedge_name.replace(',', '.').upper()
        if not wedge_name:
            return False
        elif 'AK' in wedge_name:
            # For the old designation of 5-series wedges
            series, set_width = '5', wedge_name.replace('AK', '').strip()
            UI.display('This is a 5-series wedge, designated as "AK"')
        elif 'BO' in wedge_name:
            # For the old designation of 5-series wedges
            series, set_width = '221', wedge_name.replace('AK', '').strip()
            UI.display('This is a 221-series wedge, designated as "BO"')
        elif 'TPWR' in wedge_name:
            UI.display('This is a typewriter wedge.')
            series, set_width = 'TPWR', wedge_name.replace('TPWR', '').strip()
        else:
            (series, set_width) = wedge_name.strip('SE').split('-')
        # Now get the set width
        # We now should have a wedge series and set width, as strings.
        # Convert the set width to float or enter it manually.
        try:
            set_width = float(set_width)
        except ValueError:
            # Will be set later
            prompt = 'Enter the set width as a decimal fraction: '
            set_width = UI.enter_data(prompt, float)
        # Check if it has a different pica as a base for unit calculations
        prompt = 'Old British pica (0.1667") based wedge?'
        is_brit_pica = wedge_name.endswith('E') or UI.confirm(prompt)
        # We have the wedge name, so we can look the wedge up in known wedges
        # (no need to enter the unit arrangement manually)
        unit_arrangement = wedge_arrangements.table.get(series, None)
        while unit_arrangement is None:
            # Enter manually:
            prompt = ('Enter the wedge unit values for rows 1...15 '
                      'or 1...16, separated by commas.\n')
            # Now we need to be sure that all whitespace is stripped,
            # and the value written to database is a list of integers
            try:
                unit_arrangement = [int(i.strip())
                                    for i in UI.enter_data(prompt).split(',')]
                # Display warning if the number of steps is not
                # 15 or 16 (15 is most common, 16 was used for HMN and KMN).
                if not 15 <= len(unit_arrangement) <= 16:
                    UI.display('Wedge must have 15 or 16 steps - enter again.')
                    unit_arrangement = None
            except ValueError:
                UI.display('Incorrect value - enter the values again.')

        # Now we need to adjust the arrangement...
        # Add 0 as the first item
        unit_arrangement = [0] + unit_arrangement
        # Keep filling the list until we reach 16 non-zero steps
        while len(unit_arrangement) < 17:
            unit_arrangement.append(unit_arrangement[-1])
        # We now should have a correct arrangement...
        if UI.confirm('Apply changes?'):
            self.series = series
            self.set_width = set_width
            self.is_brit_pica = is_brit_pica
            self.unit_arrangement = unit_arrangement

    def delete_from_db(self):
        """Deletes the wedge from database"""
        ans = UI.confirm('Are you sure?')
        if ans and DB.delete_wedge(self):
            UI.display('Wedge definition deleted successfully from database.')
            return True

    def save_to_db(self):
        """Stores the wedge definition in database"""
        try:
            DB.add_wedge(self)
            UI.pause('Wedge saved successfully.')
            return True
        except e.DatabaseQueryError:
            UI.display()
            UI.display_parameters(self.parameters)
            UI.pause('Cannot save the wedge - check if it is already there.')
            UI.display()
            return False

    def check_db(self):
        """Checks if the wedge is in database"""
        return DB.check_wedge(self)

    def manipulation_menu(self):
        """A menu with all operations on a wedge"""
        # Menu
        try:
            while True:
                # Keep working on a chosen diecase
                UI.display('\n')
                UI.display_parameters(self.parameters)
                UI.display('\n')
                messages = ['[E]dit wedge, [S]ave to database']
                options = {'M': e.return_to_menu,
                           'E': self.edit,
                           'S': self.save_to_db,
                           '': e.menu_level_up}
                if self.check_db():
                    options['D'] = self.delete_from_db
                    messages.append(', [D]elete from database')
                # Options constructed
                messages.append('\n[M] to return to menu; leave blank to '
                                'choose/create another wedge.')
                messages.append('\nYour choice: ')
                message = ''.join(messages)
                UI.simple_menu(message, options)()
        except e.MenuLevelUp:
            # Exit wedge manipulation menu
            return True


class SelectWedge(Wedge):
    """Selects a wedge automatically or manually"""
    def __init__(self, wedge_series=None, set_width=None):
        super().__init__()
        try:
            wedge_data = (wedge_series and set_width and
                          DB.get_wedge(wedge_series, set_width) or
                          choose_wedge() or None)
            (self.series, self.set_width, self.is_brit_pica,
             self.unit_arrangement) = wedge_data
        except (TypeError, ValueError, e.NoMatchingData, e.DatabaseQueryError):
            UI.display('Wedge choice failed. Using S5-12 instead.')


def wedge_operations():
    """Wedge operations menu for inventory management"""
    try:
        UI.display('Wedge manipulation: choose a wedge or define a new one')
        while True:
            # Choose a wedge or initialize a new one
            SelectWedge().manipulation_menu()
    except e.ReturnToMenu:
        # Exit wedge operations
        return True


def generate_wedge_collection(series='5', is_brit_pica=True):
    """Generates a collection of wedges for all set widths from 5.0 to 14.75
    for a given series (S5 by default)"""
    widths = [x + y for x in range(5, 15) for y in (0, 0.25, 0.5, 0.75)]
    for set_width in widths:
        wedge = Wedge()
        wedge.series = series
        wedge.set_width = set_width
        wedge.is_brit_pica = is_brit_pica
        unit_arrangement = [0] + [x for x in wedge_arrangements.table[series]]
        while len(unit_arrangement) < 17:
            unit_arrangement.append(unit_arrangement[-1])
        wedge.unit_arrangement = unit_arrangement
        wedge.save_to_db()
    list_wedges()
    UI.pause()


def list_wedges():
    """Lists all wedges we have."""
    data = DB.get_all_wedges()
    results = {}
    UI.display('\n' +
               'Pos.'.ljust(5) +
               'Series'.ljust(10) +
               'Set'.ljust(10) +
               'Br. pica?'.ljust(10) +
               'Unit arrangement' +
               '\n')
    for index, wedge in enumerate(data, start=1):
        # Save the wedge associated with the index
        results[index] = wedge
        # Unit arrangement is the last item of the wedge list
        unit_arrangement = wedge[-1]
        # Build a row for each wedge
        data_string = ''.join([str(field).ljust(10) for field in wedge[:-1]])
        unit_values_string = ' '.join([str(x) for x in unit_arrangement if x])
        UI.display(str(index).ljust(5) + data_string + unit_values_string)
    return results


def choose_wedge():
    """Tries to choose a wedge of given series and set width.
    If that fails, lists wedges and lets the user choose one;
    returns the wedge."""
    prompt = 'Enter number (leave blank to work with default S5-12E): '
    while True:
        try:
            UI.display('\nChoose a wedge:', end='\n')
            available_wedges = list_wedges()
            choice = UI.enter_data_or_blank(prompt, int)
            return choice and available_wedges[choice] or None
        except KeyError:
            UI.pause('Wedge number is incorrect!')
        except (e.NoMatchingData, e.DatabaseQueryError):
            UI.display('No wedges found in database.')
            return None



# -*- coding: utf-8 -*-
"""Module containing functions for diecase retrieval and parsing,
diecase storing, parameter searches.
"""
# File operations
import io
import os
import csv
# Object copying
from copy import deepcopy
# Some functions raise custom exceptions
from . import exceptions as e
# Wedge operations for several matrix-case management functions
from . import wedge_data
# Constants module
from . import constants as c
# Parsing module
from . import parsing as p
# Database backend
from . import database
# User interface
from .global_settings import UI
DB = database.Database()


class Diecase(object):
    """Diecase: matrix case attributes and operations"""
    def __init__(self):
        self.diecase_id = ''
        self.type_series = ''
        self.type_size = ''
        self.typeface_name = ''
        self.wedge = wedge_data.Wedge()
        self.layout = generate_empty_layout(15, 17)

    def show_layout(self):
        """Shows the diecase layout"""
        UI.display_diecase_layout(self)
        UI.pause()

    def edit_layout(self):
        """Edits a layout and asks if user wants to save changes"""
        # Edit the layout
        UI.edit_diecase_layout(self)

    def get_matrix(self, char='', *styles):
        """Chooses a matrix manually"""
        matrix = Matrix()
        matrix.diecase = self
        if char:
            UI.display('Enter matrix data for character: %s' % char)
        styles = styles
        prompt = 'Combination? (default: G5): '
        code_string = UI.enter_data_or_blank(prompt).upper() or 'G5'
        combination = p.parse_signals(code_string)
        # Got a list of signals
        row = p.get_row(combination)
        column = p.get_column(combination)
        # Default unit width value = determine based on wedge
        units = self.wedge.unit_arrangement[row]
        prompt = 'Unit width? (default: %s): ' % units
        units = abs(UI.enter_data_or_blank(prompt, int) or units)
        matrix.char = char
        matrix.styles = styles
        matrix.column = column
        matrix.row = row
        matrix.units = units
        return matrix

    def import_layout(self):
        """Imports a layout from file"""
        # Start with an empty layout
        new_layout = generate_empty_layout()
        # Load the layout from file
        submitted_layout = import_layout_file()
        if not submitted_layout:
            UI.pause('File does not contain a proper layout!')
            return False
        # Update the empty layout with characters read from file
        # record = (char, styles, column, row, units)
        try:
            for position in new_layout:
                for record in submitted_layout:
                    if record[2] == position[2] and record[3] == position[3]:
                        new_layout[new_layout.index(position)] = record
        except (TypeError, ValueError):
            UI.pause('Cannot process the uploaded layout!')
            return False
        # Other positions will be empty but defined
        # Now display the layout - need to use a temporary diecase for that
        temp_diecase = Diecase()
        temp_diecase.layout = new_layout
        temp_diecase.wedge = self.wedge
        UI.display('\nSubmitted layout:\n')
        UI.display_diecase_layout(temp_diecase)
        # Ask for confirmation
        if UI.confirm('Save the changes?'):
            self.layout = new_layout
            return True

    def export_layout(self):
        """Exports the matrix case layout to file."""
        filename = os.path.expanduser('~') + '/%s.csv' % self.diecase_id
        with io.open(filename, 'a') as output_file:
            csv_writer = csv.writer(output_file, delimiter=';', quotechar='"',
                                    quoting=csv.QUOTE_ALL)
            csv_writer.writerow(['Character', 'Style(s)', 'Column', 'Row',
                                 'Unit width'])
            for record in self.layout:
                (char, styles, column, row, units) = record
                csv_writer.writerow([char, ', '.join(styles),
                                     column, row, units])
        UI.pause('File %s successfully saved.' % filename)
        return True

    def clear_layout(self):
        """Clears a layout for the diecase"""
        layout = generate_empty_layout()
        if UI.confirm('Are you sure?'):
            self.layout = layout

    def set_diecase_id(self, diecase_id=None):
        """Sets a diecase ID"""
        prompt = 'Diecase ID? (leave blank to exit) : '
        diecase_id = (diecase_id or UI.enter_data_or_blank(prompt) or
                      self.diecase_id)
        # Ask if we are sure we want to update this
        # if self.diecase_id was set earlier
        condition = (not self.diecase_id or diecase_id != self.diecase_id and
                     UI.confirm('Are you sure to change diecase ID?'))
        if condition:
            self.diecase_id = diecase_id
            return True

    def set_typeface(self, type_series=None, type_size=None,
                     typeface_name=None):
        """Sets the type series, size and typeface name"""
        prompt = 'Type series: '
        type_series = (type_series or UI.enter_data_or_blank(prompt) or
                       self.type_series)
        type_size = (type_size or UI.enter_data('Type size in points: ') or
                     self.type_size)
        typeface_name = (typeface_name or UI.enter_data('Typeface name: ') or
                         self.typeface_name)
        # Validate data
        current_data_not_set = not self.type_series and not self.type_size
        if current_data_not_set or UI.confirm('Apply changes?'):
            self.type_series = type_series
            self.type_size = type_size
            self.typeface_name = typeface_name

    def assign_wedge(self, wedge_series=None, set_width=0):
        """Assigns a wedge (from database or newly-defined) to the diecase"""
        self.wedge = wedge_data.SelectWedge(wedge_series, set_width)

    @property
    def parameters(self):
        """Gets a list of parameters"""
        return [('\n', '\nMatrix case data'),
                (self.diecase_id, 'Diecase ID'),
                (self.typeface_name, 'Typeface'),
                (self.type_series, 'Type series'),
                (self.type_size, 'Type size')]

    def save_to_db(self):
        """Stores the matrix case definition/layout in database"""
        try:
            DB.add_diecase(self)
            UI.pause('Data saved successfully.')
            return True
        except e.DatabaseQueryError:
            UI.pause('Cannot save the diecase!')

    def delete_from_db(self):
        """Deletes a diecase from database"""
        ans = UI.confirm('Are you sure?')
        if ans and DB.delete_diecase(self):
            UI.display('Matrix case deleted successfully.')
            return True

    def check_db(self):
        """Checks if the diecase is registered in database"""
        try:
            DB.get_diecase(self.diecase_id)
            return True
        except (e.DatabaseQueryError, e.NoMatchingData):
            return False

    def get_styles(self):
        """Parses the diecase layout and gets available typeface styles.
        Returns a list of them."""
        try:
            return list({style for mat in self.layout
                         for style in mat[1] if style})
        except TypeError:
            return []

    def copy(self):
        """Copies itself and returns an independent object"""
        return deepcopy(self)

    def manipulation_menu(self):
        """A menu with all operations on a diecase"""
        # Menu
        try:
            while True:
                UI.clear()
                UI.display_parameters(self.parameters, self.wedge.parameters)
                options = {'M': e.return_to_menu,
                           'T': self.set_typeface,
                           'W': self.assign_wedge,
                           'ID': self.set_diecase_id,
                           'E': self.edit_layout,
                           'C': self.clear_layout,
                           'V': self.show_layout,
                           'I': self.import_layout,
                           'X': self.export_layout,
                           '': e.menu_level_up}
                messages = ['\nMatrix case manipulation:\n\n'
                            '[V]iew, [C]lear or resize, [E]dit, [I]mport '
                            'or e[X]port layout\nAssign [W]edge, '
                            'change [T]ypeface or diecase [ID]\n\n']
                # Save to database needs a complete set of metadata
                required = {'Type series': self.type_series,
                            'Type size': self.type_size,
                            'Typeface / font family name': self.typeface_name,
                            'Diecase ID': self.diecase_id}
                missing = [item for item in required if not required[item]]
                messages.extend([item + ' is not set\n' for item in missing])
                if not missing:
                    options['S'] = self.save_to_db
                    messages.append('[S]ave diecase to database')
                # Check if it's in the database
                if self.check_db():
                    options['D'] = self.delete_from_db
                    messages.append(', [D]elete diecase from database')
                # Options constructed
                messages.append('\n[M] to exit to menu, or leave blank '
                                'to choose/create another diecase.')
                messages.append('\nYour choice: ')
                message = ''.join(messages)
                UI.simple_menu(message, options)()
        except e.MenuLevelUp:
            # Exit matrix case manipulation menu
            return True


class SelectDiecase(Diecase):
    """Initialize a diecase object with or without given ID.
    If possible, choose automatically. If this fails, choose manually,
    and if user chooses no diecase, an empty one will be initialized."""
    def __init__(self, diecase_id=None):
        super().__init__()
        # Diecases created with diecase_id will be set up automatically
        try:
            diecase_data = (diecase_id and DB.get_diecase(diecase_id) or
                            choose_diecase())
            (self.diecase_id, self.type_series, self.type_size, wedge_series,
             set_width, self.typeface_name, self.layout) = diecase_data
            self.wedge = wedge_data.SelectWedge(wedge_series, set_width)
        except (KeyError, TypeError, e.NoMatchingData, e.DatabaseQueryError):
            UI.display('Diecase choice failed. Using empty one instead.')


class Matrix(object):
    """A class for single matrices - all matrix data"""
    def __init__(self):
        self.char = ''
        self.styles = []
        self.column = 'O'
        self.row = 15
        self.units = 18
        self.diecase = Diecase()

    @property
    def row_units(self):
        """Gets the row's unit width value for the diecase's wedge"""
        return self.diecase.wedge.unit_arrangement[self.row]

    @property
    def row(self):
        """Gets the row number"""
        return self.__dict__.get('_row', 5)

    @row.setter
    def row(self, value):
        """Gets the row number"""
        value = int(value)
        value = max(1, value)
        value = min(value, 16)
        self.__dict__['_row'] = value

    @property
    def code(self):
        """Gets the matrix code"""
        return self.column + str(self.row)


def diecase_operations():
    """Matrix case operations menu for inventory management"""
    try:
        UI.display('Matrix case manipulation: '
                   'choose a diecase or define a new one')
        while True:
            # Choose a wedge or initialize a new one
            SelectDiecase().manipulation_menu()
    except e.ReturnToMenu:
        # Exit wedge operations
        return True


def list_diecases():
    """Lists all matrix cases we have."""
    data = DB.get_all_diecases()
    results = {}
    UI.display('\n' +
               'Index'.ljust(7) +
               'Diecase ID'.ljust(20) +
               'Type series'.ljust(15) +
               'Type size'.ljust(15) +
               'Wedge series'.ljust(15) +
               'Set width'.ljust(15) +
               'Typeface name' + '\n')
    for index, diecase in enumerate(data, start=1):
        # Collect diecase parameters
        index = str(index)
        row = [index.ljust(7)]
        row.append(str(diecase[0]).ljust(20))
        row.extend([str(field).ljust(15) for field in diecase[1:-2]])
        # Add number and ID to the result that will be returned
        results[index] = diecase[0]
        # Add typeface name - no justification!
        row.append(diecase[-2])
        UI.display(''.join(row))
    UI.display('\n\n')
    # Now we can return the number - diecase ID pairs
    return results


def choose_diecase():
    """Lists diecases and lets the user choose one;
    returns the Diecase class object with all parameters set up."""
    prompt = 'Number of a diecase or leave blank for an empty one: '
    while True:
        try:
            UI.display('Choose a matrix case:', end='\n\n')
            data = list_diecases()
            choice = UI.enter_data_or_blank(prompt)
            return choice and DB.get_diecase(data[choice]) or None
        except KeyError:
            UI.pause('Diecase number is incorrect!')
        except (e.NoMatchingData, e.DatabaseQueryError):
            UI.display('No diecases found in database')
            return None


def import_layout_file():
    """Reads a matrix case arrangement from a text or csv file.
    The format should be:
    "character";"style1,style2...";"column";"row";"unit_width"
    """
    # Give us a file or end here
    filename = UI.enter_input_filename()
    if not filename:
        return False
    # Initialize the records list
    all_records = []
    # This will store the processed combinations - and whenever a duplicate
    # is detected, the function will raise an exception
    with io.open(filename, 'r') as layout_file:
        input_data = csv.reader(layout_file, delimiter=';', quotechar='"')
        all_records = [record for record in input_data]
        displayed_lines = [' '.join(record) for record in all_records[:5]]
        # Preview file
        UI.display('File preview: displaying first 5 rows:\n')
        UI.display('\n'.join(displayed_lines), end='\n\n')
        # Ask if the first row is a header - if so, away with it
        if UI.confirm('Is the 1st row a table header? '):
            all_records.pop(0)
    if not UI.confirm('Proceed?'):
        return False
    try:
        # Process the records
        processed_records = [process_record(record) for record in all_records]
        # Determine the diecase size based on row and column
        # Get columns and rows lists
        columns = {record[2] for record in processed_records}
        rows = sorted({record[3] for record in processed_records})
        # Check if 17 columns (15x17, 16x17), else 15 columns (old 15x15)
        big_layout = 'NI' in columns or 'NL' in columns or 16 in rows
        columns_list = big_layout and c.COLUMNS_17 or c.COLUMNS_15
        # We now have completed uploading a layout and making a list out of it
        layout = [record for row in rows for col in columns_list
                  for record in processed_records
                  if record[2] == col and record[3] == row]
        # Show the uploaded layout
        return layout
    except (KeyError, ValueError, IndexError):
        return False


def generate_empty_layout(rows=0, columns=0):
    """Generates a table of empty values for matrix case layout"""
    options = {'1': (15, 15), '2': (15, 17), '3': (16, 17)}
    if (rows, columns) not in options.values():
        prompt = "Matrix case size: 1 for 15x15, 2 for 15x17, 3 for 16x17? "
        (rows, columns) = UI.simple_menu(prompt, options)
    # Generate column numbers
    columns_list = columns == 17 and c.COLUMNS_17 or c.COLUMNS_15
    # Generate row numbers: 1...15 or 1...16
    rows_list = [num + 1 for num in range(rows)]
    # Generate an empty layout with default row unit values
    layout = [['', ['roman'], column, row, 0]
              for row in rows_list for column in columns_list]
    return layout


def process_record(record):
    """Prepares the record found in file for adding to the layout"""
    # A record is a list with all diecase data:
    # [character, (style1, style2...), column, row, units]
    # Add a character - first item; if it's a space, don't change it
    try:
        # 5 fields in a record = unit value given
        # Unit value must be converted to int
        (char, styles, column, row, units) = record
        units = int(units.strip())
    except (ValueError, AttributeError):
        # 4 fields = unit value not given
        # (or unit value cannot be converted to int)
        (char, styles, column, row) = record
        units = 0
    if char != ' ':
        char = char.strip()
    styles = [style.strip() for style in styles.split(',')]
    row = int(row.strip())
    column = column.strip()
    # Pack it again
    return (char, styles, column, row, units)



# -*- coding: utf-8 -*-
"""
casting:

A module for everything related to working on a Monotype composition caster:
-casting composed type,
-punching paper tape (ribbon) for casters without interfaces,
-casting l lines of m sorts from matrix with x, y coordinates,
-composing and casting a line of text
-testing all valves, lines and pinblock,
-calibrating the space transfer wedge,
-heating the mould up,
-sending any codes/combinations to the caster and keeping them on.
"""

# IMPORTS:
from collections import deque
# Signals parsing methods for rpi2caster
from . import parsing as p
# Custom exceptions
from . import exceptions as e
# Constants shared between modules
from . import constants as c
# Typesetting functions module
from . import typesetting_funcs
# Caster backend
from . import monotype
# Casting stats
from .casting_stats import Stats
# Globally selected UI
from .global_settings import UI
# Matrix, wedge and typesetting data models
from . import typesetting_data
from . import matrix_data
from . import wedge_data
# Constants for control sequences
GALLEY_TRIP = 'NKJS 0005 0075'
PUMP_OFF = 'NJS 0005'
PUMP_ON = 'NKS 0075'


def choose_sensor_and_driver(casting_routine):
    """Checks current modes (simulation, perforation, testing)"""
    def wrapper(self, *args, **kwargs):
        """Wrapper function"""
        UI.debug_pause('About to %s' %
                       (self.caster.mode.casting and 'cast composition...' or
                        self.caster.mode.testing and 'test outputs...' or
                        self.caster.mode.punching and 'punch ribbon...'))
        # Instantiate and enter context
        with self.caster.mode.sensor() as self.caster.sensor:
            with self.caster.mode.output() as self.caster.output:
                with self.caster:
                    return casting_routine(self, *args, **kwargs)
    return wrapper


def cast_or_punch_result(ribbon_source):
    """Ask for confirmation and cast the resulting ribbon;
    uses temporary stats object"""
    def wrapper(self, *args, **kwargs):
        """Wrapper function"""
        try:
            self.cast_queue(ribbon_source(self, *args, **kwargs))
        except e.CastingAborted:
            pass
        self.caster.mode.diagnostics = False
        self.caster.mode.hmn, self.caster.mode.unitshift = False, False
    return wrapper


def prepare_job(casting_workflow):
    """Prepares the job for casting"""
    line_of_quads = ['O15'] * 20 + ['NKJS 0005 0075']
    # Alias to make it shorter
    enter = UI.enter_data_or_blank

    def wrapper(self, ribbon):
        """Wrapper function"""
        # Mode aliases
        punching = self.caster.mode.punching
        diagnostics = self.caster.mode.diagnostics
        # Rewind the ribbon if 0005 is found before 0005+0075
        if not diagnostics and not punching and p.stop_comes_first(ribbon):
            ribbon = [x for x in reversed(ribbon)]
        self.stats.ribbon = ribbon
        if not diagnostics:
            UI.display_header('Ribbon info:')
            UI.display_parameters(self.stats.parameters)
            UI.display('\n')
        # Ask how many times to repeat this (always 1 run for calibrating)
        prompt = 'How many times do you want to cast it? (default: 1) : '
        self.stats.runs = ((diagnostics or punching) * 1 or
                           abs(enter(prompt, int) or 1))
        # Line skipping - ask user if they want to skip any initial line(s)
        prompt = 'How many lines do you want to skip?  (default: 0): '
        l_skipped = (not diagnostics and not punching and
                     abs(enter(prompt, int) or 0))
        # Tell the stats how many times we do the job
        self.stats.runs_left = self.stats.runs
        # Leave at least one line in the ribbon to cast
        l_skipped = min(l_skipped, self.stats.ribbon_lines - 1)
        if not diagnostics:
            UI.display_header('Session info:')
            UI.display_parameters(self.stats.session_parameters)
        while self.stats.runs_left:
            queue = deque(ribbon)
            quad_lines = 0
            if not diagnostics:
                UI.display_header('Starting run %s of %s, %s left'
                                  % (self.stats.current_run_number,
                                     self.stats.runs, self.stats.runs_left))
            if not (punching or diagnostics) and UI.confirm('Heat the mould?'):
                l_skipped -= 2
                # When skipping less than 2 lines, we'll cast quads instead
                quad_lines = max(-l_skipped, 0)
            # Apply constraints: 0 <= lines_skipped < lines in ribbon
            l_skipped = max(0, l_skipped)
            if l_skipped:
                UI.display('Skipping %s initial lines...' % l_skipped)
            if quad_lines:
                UI.display('Casting %s line(s) of 20 quads' % quad_lines)
            # Make sure we take off lines from the beginning of the
            # casting job - i.e. last lines of text; count the lines
            code = ''
            # Add one more line to skip - because ribbon starts with 0075-0005
            l_skipped += 1
            while l_skipped and not diagnostics:
                code = queue.popleft()
                l_skipped -= 1 * p.check_newline(code)
            # Add it back, then add those quads
            queue.appendleft(code)
            queue.extendleft(quad_lines * line_of_quads)
            # The ribbon is ready for casting / punching
            try:
                casting_workflow(self, queue or [])
                if not self.stats.runs_left and UI.confirm('Repeat?'):
                    self.stats.add_run()
            except e.CastingAborted:
                l_skipped = self.stats.current_line_number
                if not UI.confirm('Retry this job?'):
                    return
    return wrapper


class Casting(object):
    """Casting:

    Methods related to operating the composition caster.
    Requires configured caster.

    All methods related to operating a composition caster are here:
    -casting composition and sorts, punching composition,
    -calibrating the caster,
    -testing the interface,
    -sending an arbitrary combination of signals,
    -casting spaces to heat up the mould."""

    def __init__(self, ribbon_file=''):
        # Caster for this job
        self.caster = monotype.MonotypeCaster()
        self.stats = Stats(self)
        self.ribbon = (ribbon_file and
                       typesetting_data.SelectRibbon(filename=ribbon_file) or
                       typesetting_data.Ribbon())

    @choose_sensor_and_driver
    @prepare_job
    def cast_queue(self, casting_queue):
        """Casts the sequence of codes in ribbon or self.ribbon.contents,
        displaying the statistics (depending on context:
        casting, punching or testing)
        """
        self.stats.next_run()
        self.stats.queue = casting_queue
        if not self.caster.mode.diagnostics:
            UI.display_header('Current run info:')
            UI.display_parameters(self.stats.run_parameters)
            UI.display('\n')
        # Now process the queue
        generator = (p.parse_record(record) for record in casting_queue)
        if not self.caster.mode.testing:
            self.caster.sensor.check_if_machine_is_working()
        for (signals, comment) in generator:
            UI.display('\n\n')
            if comment:
                UI.display(comment + '\n' + '-' * len(comment))
            if not signals:
                # No signals (comment only)- go to the next combination
                continue
            # Check if HMN or unit-shift must be applied
            self.stats.signals = signals
            UI.display_parameters(self.stats.code_parameters)
            # Let the caster do the job
            self.caster.process(signals)
        return True

    @cast_or_punch_result
    def _test_front_pinblock(self):
        """Sends signals 1...14, one by one"""
        UI.pause('Testing the front pinblock - signals 1 towards 14.')
        self.caster.mode.testing = True
        return [str(n) for n in range(1, 15)]

    @cast_or_punch_result
    def _test_rear_pinblock(self):
        """Sends NI, NL, A...N"""
        UI.pause('This will test the front pinblock - signals NI, NL, A...N. ')
        self.caster.mode.testing = True
        return [x for x in c.COLUMNS_17]

    @cast_or_punch_result
    def _test_all(self):
        """Tests all valves and composition caster's inputs in original
        Monotype order: NMLKJIHGFSED 0075 CBA 123456789 10 11 12 13 14 0005.
        """
        UI.pause('This will test all the air lines in the same order '
                 'as the holes on the paper tower: \n%s\n'
                 'MAKE SURE THE PUMP IS DISENGAGED.' % ' '.join(c.SIGNALS))
        self.caster.mode.testing = True
        return [x for x in c.SIGNALS]

    @cast_or_punch_result
    def _test_justification(self):
        """Tests the 0075-S-0005"""
        UI.pause('This will test the justification pin block (0075, S, 0005).')
        self.caster.mode.testing = True
        return ['0075', 'S', '0005']

    @choose_sensor_and_driver
    def _test_any_code(self):
        """Tests a user-specified combination of signals"""
        self.caster.mode.testing = False
        while True:
            UI.display('Enter the signals to send to the caster, '
                       'or leave empty to return to menu: ')
            string = UI.enter_data_or_blank('Signals? (leave blank to exit): ')
            signals = p.parse_signals(string)
            if signals:
                UI.display('Activating ' + ' '.join(signals))
                self.caster.output.valves_on(signals)
            else:
                break
        self.caster.output.valves_off()
        self.caster.mode.testing = False

    @cast_or_punch_result
    def cast_composition(self):
        """Casts or punches the ribbon contents"""
        ribbon = self.ribbon.contents
        return ribbon and ribbon or e.return_to_menu()

    @cast_or_punch_result
    def cast_sorts(self, source=()):
        """Sorts casting routine, based on the position in diecase.
        Ask user about the diecase row & column, as well as number of sorts.
        """
        queue = []
        casting_queue = [x for x in source] or [(None, None, 0)]
        for (char, style, qty) in casting_queue:
            # Make an infinite iterator when we call this function without
            # the list of character-style pairs
            prompt = ('Character: %s\n'
                      'Y: add to casting queue, N: skip it?' % char)
            if not char:
                casting_queue.append((char, style, 0))
            elif not UI.confirm(prompt):
                continue
            prompt = 'Character? (leave blank to end specifying characters): '
            char = char or UI.enter_data_or_blank(prompt)
            if not char:
                break
            available_styles = {'r': 'roman', 'b': 'bold',
                                'i': 'italic', 's': 'smallcaps',
                                'l': 'subscript', 'u': 'superscript'}
            style = style or (available_styles.get(UI.enter_data_or_blank(
                'Styles: [r]oman, [b]old, [i]talic, [s]mall caps, '
                '[l]ower index, [u]pper index (default: roman): ', str), 'r'))
            matrix = self.diecase.get_matrix(char, [style])
            diff = matrix.units - matrix.row_units
            wedge_positions = self._calculate_wedges(diff)
            # Ask for number of sorts and lines, no negative numbers here
            prompt = '\nHow many sorts per line? (default: 10): '
            qty = qty or abs(UI.enter_data_or_blank(prompt, int) or 10)
            prompt = '\nHow many lines? (default: 1): '
            lines = abs(UI.enter_data_or_blank(prompt, int) or 1)
            UI.display('\n')
            # Warn if we want to cast too many sorts from a single matrix
            warning = ('Warning: you want to cast a single character more than'
                       ' 10 times. This may lead to matrix overheating!\n')
            if qty > 10:
                UI.display(warning)
            UI.display('\n')
            line_codes = [GALLEY_TRIP + str(wedge_positions['0005']),
                          PUMP_ON + str(wedge_positions['0075'])] + ['O15'] * 2
            # 2 quads in the beginning, then spaces, then 2 quads in the end
            line_codes.extend([matrix.code + (diff and 'S' or '') +
                               '// ' + matrix.char] * qty)
            line_codes.extend(['O15'] * 2)
            # Ask for confirmation
            queue.extend(line_codes * lines)
        # No sequences? No casting.
        if queue and UI.confirm('Cast the sequence?'):
            UI.display('\nEach line will have two em-quads at the start '
                       'and at the end, to support the type.\n')
            return queue + [GALLEY_TRIP, PUMP_OFF, PUMP_OFF]
        else:
            e.return_to_menu()

    @cast_or_punch_result
    def cast_spaces(self):
        """Spaces casting routine, based on the position in diecase.
        Ask user about the space width and measurement unit.
        """
        queue = []
        # Measurement system
        prompt = ('Measurement unit to use? \n'
                  '[A]merican Johnson pica = 0.166",'
                  '[B]ritish pica = 0.1667",\n[D]idot cicero = 0.1776", '
                  '[F]ournier cicero = 0.1629": ')
        pica = UI.simple_menu(prompt, {'A': 0.1660, 'B': 0.1667,
                                       'D': 0.1776, 'F': 0.1629})
        # Ask about line length
        prompt = 'Galley line length in above units? (default: 25) : '
        line_length = abs(UI.enter_data_or_blank(prompt, int) or 25)*1.0
        # We need 72 additional units for two quads before and after line
        line_length = int(18 * pica / 0.1667 * line_length *
                          self.wedge.set_width / 12) - 72
        while True:
            matrix = self.diecase.get_matrix(' ')
            prompt = ('Space width? [6] = 1/6em, [4] = 1/4em, [3] = 1/3em, '
                      '[2] = 1/2em, [1] = 1em, [C] for custom width: ')
            # Width in points
            width = UI.simple_menu(prompt, {'6': 1/6, '4': 1/4, '3': 1/3,
                                            '2': 1/2, '1': 1, 'C': 0}) * 12.0
            # Ask about custom value, then specify units
            while not 2 <= width <= 20:
                prompt = 'Custom width in points (decimal, 2...20) ? : '
                width = UI.enter_data_or_blank(prompt, float)
            prompt = '\nHow many lines? (default: 1): '
            lines = abs(UI.enter_data_or_blank(prompt, int) or 1)
            prompt = ('Casting %s lines of %s-point spaces from %s%s. OK?'
                      % (lines, width, matrix.column, matrix.row))
            # Save for later
            comment = '%s-point space' % width
            # Repeat
            if not UI.confirm(prompt):
                continue
            # Calculate unit width of said number of points
            # We do it this way:
            # units = picas * set_width/12 * 18
            # a pica is 12 points, so:
            # units = points * set_width/12 * 1 / 12 * 18
            # 18 / (12*12) = 0.125, hence division by 8
            width = round(width * pica / self.wedge.pica *
                          self.wedge.set_width / 8, 2)
            # How many spaces will fit in a line? Calculate it...
            qty = int(line_length // width)
            # Check if the corrections are needed at all
            diff = width - matrix.row_units
            wedge_positions = self._calculate_wedges(diff)
            # Add 'S' if there is width difference
            signals = matrix.code + (diff and 'S' or '') + '// ' + comment
            line_codes = [GALLEY_TRIP + str(wedge_positions['0005']),
                          PUMP_ON + str(wedge_positions['0075'])]
            # 2 quads in the beginning, then spaces, then 2 quads in the end
            line_codes.extend(['O15'] * 2 + [signals] * qty + ['O15'] * 2)
            # Ask for confirmation
            queue.extend(line_codes * lines)
            if queue and UI.confirm('Y to cast, N to add some more spaces?'):
                return queue + [GALLEY_TRIP, PUMP_OFF, PUMP_OFF]

    def cast_typecases(self):
        """Casting typecases according to supplied font scheme."""
        scheme = typesetting_data.SelectFontScheme()
        prompt = ('Scale in %, relative to 100 "a" characters:\n'
                  '(leave blank to exit) ?: ')
        scale = UI.enter_data_or_blank(prompt, float) / 100
        if not scale:
            return
        available_styles = {'r': 'roman', 'b': 'bold',
                            'i': 'italic', 's': 'smallcaps',
                            'l': 'subscript', 'u': 'superscript'}
        print('Styles to cast?\n'
              'Available styles: [r]oman, [b]old, [i]talic, [s]mall caps,\n'
              '[l]ower index (a.k.a. subscript, inferior), '
              '[u]pper index (a.k.a. superscript, superior).\n'
              'Leave blank for roman.')
        styles_string = UI.enter_data_or_blank('Styles?: ') or 'r'
        styles = [available_styles.get(char, None) for char in styles_string]
        styles = [style for style in styles if style is not None]
        # Now generate and display the casting queue
        sorts = {style: [(char, style, round(qty * scale))
                         for (char, ch_style, qty) in scheme.layout
                         if ch_style == style] for style in styles}
        queue = []
        for style in sorts:
            UI.display_header(style)
            queue.extend(sorts[style])
            for (char, _, qty) in sorts[style]:
                UI.display('%s : %s' % (char, qty))
        self.cast_sorts(queue)

    @cast_or_punch_result
    def adhoc_typesetting(self):
        """Allows us to use caster for casting single lines.
        This means that the user enters a text to be cast,
        gives the line length, chooses alignment and diecase.
        Then, the program composes the line, justifies it, translates it
        to Monotype code combinations.

        This allows for quick typesetting of short texts, like names etc.
        """
        # Initialize the typesetter for a chosen diecase
        typesetter = typesetting_funcs.Typesetter()
        # Supply the diecase id
        typesetter.session_setup(self.diecase)
        # Enter text
        text = UI.enter_data("Enter text to compose: ")
        typesetter.text_source = typesetter.parse_and_generate(text)
        # Translate the text to Monotype signals
        typesetter.compose()
        queue = typesetter.justify()
        self.ribbon.contents, old_ribbon = queue, self.ribbon.contents
        if UI.confirm('Show the codes?'):
            self.ribbon.display_contents()
        self.ribbon.contents = old_ribbon
        return queue

    @cast_or_punch_result
    def _calibrate_wedges(self):
        """Allows to calibrate the justification wedges so that when you're
        casting a 9-unit character with the S-needle at 0075:3 and 0005:8
        (neutral position), the    width is the same.

        It works like this:
        1. 0075 - turn the pump on,
        2. cast 10 spaces from the specified matrix (default: G9),
        3. put the line to the galley & set 0005 to 8, 0075 to 3, pump on,
        4. cast 10 spaces with the S-needle from the same matrix,
        5. put the line to the galley, then 0005 to turn the pump off.
        """
        intro = ('Transfer wedge calibration:\n\n'
                 'This function will cast two lines of 5 spaces: '
                 'first: G5, second: GS5 with wedges at 3/8. \n'
                 'Adjust the 52D space transfer wedge '
                 'until the lengths are the same.')
        UI.display(intro)
        signals = 'G5'
        queue = [GALLEY_TRIP] + [signals] * 7
        queue.extend([GALLEY_TRIP + '8', PUMP_ON + '3'] + [signals + 'S'] * 7)
        queue.extend([GALLEY_TRIP, PUMP_OFF, PUMP_OFF])
        self.caster.mode.calibration = True
        return queue

    @cast_or_punch_result
    def _calibrate_mould(self):
        """Calculates the width, displays it and casts some 9-unit characters.
        Then, the user measures the width and adjusts the mould opening width.
        """
        intro = ('Mould blade opening calibration:\n'
                 'Cast G5 (9-units wide on S5 wedge), then measure the width. '
                 'Adjust if needed.')
        UI.display(intro)
        # We calculate the width of double 9 units = 18 units, i.e. 1 pica em
        em_width = self.wedge.pica * self.wedge.set_width / 12.0
        UI.display('9 units (1en) is %s" wide' % round(0.5 * em_width, 4))
        UI.display('18 units (1em) is %s" wide' % round(em_width, 4))
        signals = 'G5'
        self.caster.mode.calibration = True
        return [GALLEY_TRIP] + [signals] * 7 + [GALLEY_TRIP] + [PUMP_OFF] * 2

    def _calibrate_diecase(self):
        """Casts the "en dash" characters for calibrating the character X-Y
        relative to type body."""
        UI.display('X-Y character calibration:\n'
                   'Cast some en-dashes and/or lowercase "n" letters, '
                   'then check the position of the character relative to the '
                   'type body.\nAdjust if needed.')
        self.caster.mode.calibration = True
        self.cast_sorts([('–', 'roman', 7), ('n', 'roman', 7),
                         ('h', 'roman', 7)])

    def diagnostics_submenu(self):
        """Settings and alignment menu for servicing the caster"""
        def menu_options():
            """Build a list of options, adding an option if condition is met"""
            caster = not self.caster.mode.punching
            opts = [(e.menu_level_up, 'Back',
                     'Returns to main menu', True),
                    (self._test_all, 'Test outputs',
                     'Tests all the air outputs one by one', True),
                    (self._test_front_pinblock, 'Test the front pinblock',
                     'Tests the pins 1...14', caster),
                    (self._test_rear_pinblock, 'Test the rear pinblock',
                     'Tests the pins NI, NL, A...N one by one', caster),
                    (self._test_justification, 'Test the 0075-S-0005 pinblock',
                     'Tests the pins for justification wedges', caster),
                    (self._test_any_code, 'Send specified signals',
                     'Sends the specified signal combination', True),
                    (self._calibrate_wedges, 'Calibrate the 52D wedge',
                     'Calibrate the space transfer wedge for correct width',
                     caster),
                    (self._calibrate_mould, 'Calibrate mould opening',
                     'Casts 9-unit characters to adjust the type width',
                     caster),
                    (self._calibrate_diecase, 'Calibrate matrix X-Y',
                     'Calibrate the character-to-body positioning', caster)]
            return [(function, description, long_description) for
                    (function, description, long_description, condition)
                    in opts if condition]

        header = ('Diagnostics and machine calibration menu:\n\n')
        # Keep displaying the menu and go back here after any method ends
        while True:
            try:
                # Catch "return to menu" and "exit program" exceptions here
                UI.menu(menu_options(), header=header)()
            except e.ReturnToMenu:
                # Stay in the menu
                pass

    def _main_menu_options(self):
        """Build a list of options, adding an option if condition is met"""
        # Options are described with tuples: (function, description, condition)
        caster = not self.caster.mode.punching
        ribbon = self.ribbon.contents
        diecase_selected = self.diecase.diecase_id
        opts = [(e.exit_program, 'Exit', 'Exits the program', True),
                (self.cast_composition, 'Cast or punch composition',
                 'Casts type or punch a ribbon', ribbon),
                (self._choose_ribbon, 'Select ribbon',
                 'Selects a ribbon from database or file', True),
                (self._choose_diecase, 'Select diecase',
                 'Selects a matrix case from database', True),
                (self._choose_wedge, 'Select wedge',
                 'Selects a wedge from database', True),
                (self.ribbon.display_contents, 'View codes',
                 'Displays all sequences in a ribbon', ribbon),
                (self.diecase.show_layout, 'Show diecase layout',
                 'Views the matrix case layout', diecase_selected),
                (self.adhoc_typesetting, 'Ad-hoc typesetting',
                 'Compose and cast a line of text', diecase_selected),
                (self.cast_sorts, 'Cast sorts',
                 'Cast from matrix with given coordinates', caster),
                (self.cast_spaces, 'Cast spaces or quads',
                 'Casts spaces or quads of a specified width', caster),
                (self.cast_typecases, 'Cast typecases',
                 'Casts a typecase based on a selected font scheme',
                 caster and diecase_selected),
                (self._display_details, 'Show detailed info...',
                 'Displays caster, ribbon, diecase and wedge details', True),
                (self.diagnostics_submenu, 'Service...',
                 'Interface and machine diagnostic functions', True)]
        # Built a list of menu options conditionally
        return [(function, description, long_description)
                for (function, description, long_description, condition)
                in opts if condition]

    def main_menu(self):
        """Main menu for the type casting utility."""
        header = ('rpi2caster - CAT (Computer-Aided Typecasting) '
                  'for Monotype Composition or Type and Rule casters.\n\n'
                  'This program reads a ribbon (from file or database) '
                  'and casts the type on a composition caster.'
                  '\n\nCasting / Punching Menu:')
        # Keep displaying the menu and go back here after any method ends
        while True:
            # Catch any known exceptions here
            try:
                UI.menu(self._main_menu_options(), header=header, footer='')()
            except (e.ReturnToMenu, e.MenuLevelUp):
                # Will skip to the end of the loop, and start all over
                pass

    @property
    def stats(self):
        """Ribbon/job statistics"""
        return self.__dict__.get('_stats', Stats(self))

    @stats.setter
    def stats(self, stats):
        """Parse the stats to get the ribbon data"""
        self.__dict__['_stats'] = (isinstance(stats, Stats) and stats or
                                   Stats(self))

    @property
    def ribbon(self):
        """Ribbon for the casting session"""
        return self.__dict__.get('_ribbon', typesetting_data.Ribbon())

    @ribbon.setter
    def ribbon(self, ribbon):
        """Ribbon setter"""
        rib = (isinstance(ribbon, typesetting_data.Ribbon) and ribbon or
               typesetting_data.SelectRibbon())
        self.__dict__['_ribbon'] = rib
        self.diecase = self.ribbon.diecase

    @property
    def diecase(self):
        """Ribbon for the casting session"""
        return self.__dict__.get('_diecase', matrix_data.Diecase())

    @diecase.setter
    def diecase(self, diecase):
        """Ribbon setter"""
        case = (isinstance(diecase, matrix_data.Diecase) and diecase or
                matrix_data.SelectDiecase())
        self.__dict__['_diecase'] = case
        self.wedge = self.diecase.wedge

    @property
    def wedge(self):
        """Ribbon for the casting session"""
        return self.__dict__.get('_wedge', wedge_data.Wedge())

    @wedge.setter
    def wedge(self, wedge):
        """Ribbon setter"""
        self.__dict__['_wedge'] = (isinstance(wedge, wedge_data.Wedge) and
                                   wedge or wedge_data.SelectWedge())

    def _choose_ribbon(self):
        """Chooses a ribbon from database or file"""
        self.ribbon = typesetting_data.SelectRibbon()

    def _choose_diecase(self):
        """Chooses a diecase from database"""
        self.diecase = matrix_data.SelectDiecase()

    def _choose_wedge(self):
        """Chooses a wedge from registered ones"""
        self.wedge = wedge_data.SelectWedge()

    def _display_details(self):
        """Collect ribbon, diecase and wedge data here"""
        UI.display_parameters(self.caster.parameters, self.ribbon.parameters,
                              self.diecase.parameters, self.wedge.parameters)
        UI.pause()

    def _calculate_wedges(self, diff):
        """calculate_wedges:

        Calculates and returns wedge positions for character.
        Uses pre-calculated unit width difference between row's unit width
        and character's width (with optional corrections).
        """
        # Calculate the inch width of delta
        # 1 pica = 18 units 12 set = 0.1667 (old British pica)
        # or 0.1660 (Am. pica); unit_width = 12 * pica / (set width * 18)
        steps_0005 = int(diff * 0.166 / self.wedge.pica *
                         self.wedge.set_width * 2000 / 1296) + 53
        # Adjust the wedges
        # You do it in respect to the neutral position i.e. 3/8:
        # 3 steps of 0075 and 8 steps of 0005 wedge.
        # 3 / 8 = 1 * 15 + 8 = 53 - that would be 53 steps of 0005 wedge
        # Add safeguards against wrong wedge positions
        # Minimum wedge positions (0075/0005) are 1/1, maximum 15/15
        # This is equivalent to min 1*15+1=16 steps and max 15*15+15=240 steps
        # WARNING - the actual value when casting chars/high spaces must not be
        # wider than the matrix (.2" for small composition, 0.4 for 2x2 etc),
        # otherwise we get a splash between mats!
        steps_0005 = min(steps_0005, 240)
        # Lower limit: 1/1 wedge positions:
        steps_0005 = max(16, steps_0005)
        steps_0075 = 0
        while steps_0005 > 15:
            steps_0005 -= 15
            steps_0075 += 1
        # Got the wedge positions, return them
        return {'0075': steps_0075, '0005': steps_0005}



# -*- coding: utf-8 -*-
"""Casting stats - implements stats getting and setting"""
from . import parsing as p


class Stats(object):
    """Casting statistics gathering and displaying functions"""
    def __init__(self, session):
        self.__dict__['_current'] = {}
        self.__dict__['_previous'] = {}
        self.__dict__['_run_data'] = {}
        self.session = session
        self.ribbon = session.ribbon.contents

    def next_run(self):
        """Updates the info about the runs"""
        self.__dict__['_current']['run'] = self.current_run_number + 1
        self.__dict__['_runs_left'] = self.runs_left - 1

    def next_line(self):
        """Updates the info about the runs"""
        self.__dict__['_current']['line'] = self.current_line_number + 1
        self.__dict__['_run_data']['lines_left'] = self.lines_left - 1

    def next_char(self):
        """Updates the characters info"""
        self.__dict__['_current']['char'] = self.current_char_number + 1

    def add_run(self):
        """Adds one to all_runs number - in case of repeating the job"""
        self.runs += 1
        self.runs_left += 1

    @property
    def parameters(self):
        """Gets ribbon parameters"""
        source = self.__dict__.get('_run_data', {})
        return [('\n', '\nRibbon statistics'),
                (source.get('codes', 0), 'Combinations in ribbon'),
                (source.get('lines', 0), 'Lines to cast'),
                (source.get('chars', 0), 'Characters incl. spaces')]

    @property
    def session_parameters(self):
        """Displays the ribbon data"""
        source = self.__dict__.get('_run_data', {})
        runs = self.runs
        return [(runs, 'Casting runs'),
                (source.get('codes', 0) * runs, 'All codes'),
                (source.get('chars', 0) * runs, 'All chars'),
                (source.get('lines', 0) * runs, 'All lines')]

    @property
    def run_parameters(self):
        """Displays info at the beginning of each run"""
        source = self.__dict__.get('_run_data', {})
        data = []
        if not self.session.caster.mode.testing:
            data.append((source.get('run_codes', 0), 'Codes in the run'))
        if self.session.caster.mode.casting:
            data.append((source.get('run_chars', 0), 'Chars in the run'))
            data.append((source.get('run_lines', 0), 'Lines in the run'))
        return data

    @property
    def code_parameters(self):
        """Displays info about the current combination"""
        # General stats for all modes:
        current = self.__dict__.get('_current', {})
        new_line = p.check_newline(current.get('signals', []))
        char = p.check_character(current.get('signals', []))
        data = [(new_line and self.current_line_number, 'Starting a new line')]
        data.append((new_line and self.lines_left, 'Lines left'))
        data.append((' '.join(current.get('signals', [])), 'Combination'))
        data.append((current.get('code', []), 'Current code'))
        pump = {True: 'ON', False: 'OFF'}[current.get('pump_status', False)]
        # For casting and punching:
        if not self.session.caster.mode.diagnostics:
            data.append((char and self.current_char_number, 'Current char'))
        if self.session.caster.mode.casting:
            data.append((current.get('0075', '15'), 'Wedge 0075 now at'))
            data.append((current.get('0005', '15'), 'Wedge 0005 now at'))
            data.append((pump, 'Pump is'))
        return data

    @property
    def runs(self):
        """Gets the runs (repetitions) number"""
        return self.__dict__.get('_runs', 1)

    @runs.setter
    def runs(self, runs=1):
        """Sets the runs (repetitions) number"""
        self.__dict__['_runs'] = isinstance(runs, int) and runs or 1

    @property
    def runs_left(self):
        """Gets the runs (repetitions) number"""
        return self.__dict__.get('_runs_left', 1)

    @runs_left.setter
    def runs_left(self, runs=1):
        """Gets the runs (repetitions) number"""
        self.__dict__['_runs_left'] = isinstance(runs, int) and runs or 1

    @property
    def lines_left(self):
        """Gets the lines number"""
        return self.__dict__.get('_run_data', {}).get('lines_left', 1)

    @property
    def current_char_number(self):
        """Gets the current character number"""
        return self.__dict__.get('_current', {}).get('char', 0)

    @property
    def current_run_number(self):
        """Gets the current run number"""
        return self.__dict__.get('_current', {}).get('run', 1)

    @property
    def current_line_number(self):
        """Gets the current line number"""
        return self.__dict__.get('_current', {}).get('line', 0)

    @property
    def ribbon_lines(self):
        """Gets number of lines per ribbon"""
        return self.__dict__.get('_run_data', {}).get('lines', 1)

    @code_parameters.setter
    def signals(self, signals):
        """Updates the stats based on current combination"""
        # Save previous state
        current_dict = self.__dict__.get('_current', {})
        self.__dict__['_previous'] = {k: v for k, v in current_dict.items()}
        # Update combination info
        current_dict['signals'] = signals
        current_dict['code'] = current_dict.get('code', 0) + 1
        if p.check_character(signals):
            self.next_char()
        elif p.check_newline(signals):
            self.next_line()
        # Check the pump working/non-working status in the casting mode
        if self.session.caster.mode.casting:
            self._update_wedge_positions()
            self._check_pump()

    @session_parameters.setter
    def ribbon(self, ribbon):
        """Parses the ribbon, counts combinations, lines and characters"""
        mode = self.session.caster.mode
        run_data = self.__dict__['_run_data']
        run_data['codes'] = 0
        run_data['lines'] = 0
        run_data['chars'] = 0
        # Reset the character stats
        self.__dict__['_current'] = {}
        self.__dict__['_previous'] = {}
        generator = (p.parse_record(x) for x in ribbon)
        for (signals, _) in generator:
            if signals:
                # Guards against empty combination i.e. line with comment only
                run_data['codes'] += 1
            if '16' in signals and not mode.hmn and not mode.unitshift:
                mode.choose_row16_addressing()
            if p.check_newline(signals):
                run_data['lines'] += 1
            elif p.check_character(signals):
                run_data['chars'] += 1

    @run_parameters.setter
    def queue(self, queue):
        """Parses the ribbon, counts combinations, lines and characters"""
        run_data = self.__dict__['_run_data']
        run_data['run_codes'] = 0
        run_data['run_lines'] = 0
        run_data['run_chars'] = 0
        generator = (p.parse_record(x) for x in queue)
        for (combination, _) in generator:
            if combination:
                # Guards against empty combination i.e. line with comment only
                run_data['run_codes'] += 1
            if p.check_newline(combination):
                run_data['run_lines'] += 1
            elif p.check_character(combination):
                run_data['run_chars'] += 1
        run_data['lines_left'] = run_data['run_lines'] + 1

    def _check_pump(self):
        """Checks pump based on current and previous combination"""
        prev_code = self.__dict__.get('_previous', {}).get('signals', [])
        current_code = self.__dict__.get('_current', {}).get('signals', [])
        # Was it running until now? Get it from the pump object
        running = self.session.caster.pump.is_working
        # Was it started before (0075 with or without 0005)?
        started = p.check_0075(prev_code)
        # Was it stopped (0005 without 0075)
        stopped = p.check_0005(prev_code) and not p.check_0075(prev_code)
        # Is 0005 or 0075 in current combination? (if so - temporary stop)
        on_hold = p.check_0005(current_code) or p.check_0075(current_code)
        # Determine the current status
        pump_status = (running or started) and not stopped and not on_hold
        current = self.__dict__.get('_current', {})
        pump = self.session.caster.pump
        current['pump_status'] = pump_status
        # Feed it back to pump object
        pump.is_working = current.get('pump_status', False)
        pump.current_0005 = current.get('0005', '15')
        pump.current_0075 = current.get('0075', '15')

    def _update_wedge_positions(self):
        """Gets current positions of 0005 and 0075 wedges"""
        combination = self.__dict__.get('_current', {}).get('signals', [])
        # Check 0005
        if p.check_0005(combination):
            candidates = [x for x in range(15) if str(x) in combination]
            self.__dict__['_current']['0005'] = (candidates and
                                                 str(max(candidates)) or '15')
        # Check 0075
        if p.check_0075(combination):
            candidates = [x for x in range(15) if str(x) in combination]
            self.__dict__['_current']['0075'] = (candidates and
                                                 str(max(candidates)) or '15')



# -*- coding: utf-8 -*-
"""
Database:

Database-related classes for rpi2caster suite.
"""
# IMPORTS:
# Used for serializing lists stored in database, and for communicating
# with the web application (in the future):
import json
# this module uses sqlite3 database for storing caster, interface,
# wedge, diecase & matrix parameters:
import sqlite3
# Custom exceptions
from . import exceptions
# Get global settings
from . import global_settings
# Package constants
from .constants import DEFAULT_DATABASE_PATHS


class Singleton(type):
    """Make only one object"""
    instance = None

    def __call__(cls, *args, **kw):
        if not cls.instance:
            cls.instance = super(Singleton, cls).__call__(*args, **kw)
        return cls.instance


class Database(object):
    """Database(database_path, confFilePath):

    A class containing all methods related to storing, retrieving
    and deleting data from a SQLite3 database used for config.

    We're using database because it's easy to access and modify with
    third-party programs (like sqlite, sqlitebrowser or a Firefox plugin),
    and there will be lots of data to store: diecase (matrix case)
    properties, diecase layouts, wedge unit values, caster and interface
    settings (although we may move them to config files - they're "system"
    settings best left default, instead of "foundry" settings the user has
    to set up before being able to cast, based on their type foundry's
    inventory, which varies from one place to another).

    Methods here are for reading/writing data for diecases, matrices,
    wedges (and casters & interfaces) from/to designated sqlite3 database.

    Default database path is /var/lib/rpi2caster - but you can
    override it by instantiating this class with a different name
    passed as an argument. It is necessary that the user who's running
    this program for setup has write access to the database file.
    """
    __metaclass__ = Singleton

    def __init__(self):
        database_paths = ([global_settings.DATABASE_PATH] +
                          DEFAULT_DATABASE_PATHS)
        # Connect to the database
        for path in database_paths:
            try:
                self.db_connection = sqlite3.connect(path)
                # End on first successful find
                break
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Continue looping
                pass
        else:
            # Fell off the end of the loop
            raise exceptions.WrongConfiguration('Cannot connect to '
                                                'database file %s' % path)

    def __enter__(self):
        return self

    def add_wedge(self, wedge):
        """Registers a wedge in our database."""
        # data - a list with wedge parameters to be written,
        # boolean is_brit_pica must be converted to int,
        # a unit arrangement is a JSON-encoded list
        data = [wedge.series, wedge.set_width, int(wedge.is_brit_pica),
                json.dumps(wedge.unit_arrangement)]
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                # Create the table first:
                sql = ('CREATE TABLE IF NOT EXISTS wedges ('
                       'wedge_series TEXT NOT NULL, '
                       'set_width REAL NOT NULL, '
                       'brit_pica INTEGER NOT NULL, '
                       'unit_arrangement TEXT NOT NULL, '
                       'PRIMARY KEY (wedge_series, set_width, brit_pica))')
                cursor.execute(sql)
                # Then add an entry:
                cursor.execute('INSERT INTO wedges ('
                               'wedge_series, set_width, '
                               'brit_pica, unit_arrangement'
                               ') VALUES (?, ?, ?, ?)''', data)
                self.db_connection.commit()
                return True
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def check_wedge(self, wedge):
        """Checks if a wedge with given parameters is in database"""
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                sql = ('SELECT * FROM wedges WHERE wedge_series = ?'
                       'AND set_width = ? AND brit_pica = ?')
                data = [wedge.series, wedge.set_width, int(wedge.is_brit_pica)]
                cursor.execute(sql, data)
                data = cursor.fetchall()
                # Empty list = False
                return bool(data)
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                return False

    def get_wedge(self, wedge_series, set_width):
        """get_wedge(wedge_series, set_width):

        Looks up a wedge with given ID and set width in database.
        If found, returns (series, set_width, brit_pica, unit_arrangement)

        series - string (e.g. S5) - wedge name
        set_width - float (e.g. 9.75) - set width,
        brit_pica - bool - whether this is an old-pica ("E") wedge or not,
        unit_arrangement - list of unit values for all wedge's steps.

        Else, function returns False.
        """
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                cursor.execute('SELECT * FROM wedges '
                               'WHERE wedge_series = ? AND set_width = ?',
                               [str(wedge_series), float(set_width)])
                (wedge_series, set_width, is_brit_pica,
                 raw_unit_arrangement) = cursor.fetchone()
                # Change return value of steps to list:
                wedge = [wedge_series, set_width, bool(is_brit_pica),
                         [int(x) for x in json.loads(raw_unit_arrangement)]]
                return wedge
            except (TypeError, ValueError, IndexError):
                # No data or cannot process it
                raise exceptions.NoMatchingData
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def delete_wedge(self, wedge):
        """Deletes a wedge from the database."""
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                sql = ('DELETE FROM wedges WHERE wedge_series = ? '
                       'AND set_width = ? AND brit_pica = ?')
                data = [wedge.series, wedge.set_width, wedge.is_brit_pica]
                cursor.execute(sql, data)
                return True
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def get_all_wedges(self):
        """Gets all wedges stored in database."""
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                cursor.execute('SELECT * FROM wedges')
                # Check if we got any:
                all_wedges = cursor.fetchall()
                processed_wedges = []
                if not all_wedges:
                    raise exceptions.NoMatchingData
                for record in all_wedges:
                    (wedge_series, set_width,
                     is_brit_pica, unit_arrangement) = record
                    record = [wedge_series, set_width, bool(is_brit_pica),
                              [x for x in json.loads(unit_arrangement)]]
                    processed_wedges.append(record)
                return processed_wedges
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def get_all_diecases(self):
        """Gets all diecases stored in database."""
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                cursor.execute('SELECT * FROM matrix_cases')
                # Check if we got any:
                results = cursor.fetchall()
                if not results:
                    raise exceptions.NoMatchingData
                return results
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def add_diecase(self, diecase):
        """Registers a diecase in our database."""
        # data - a list with diecase parameters to be written,
        # layout is a JSON-dumped dictionary
        data = [diecase.diecase_id, diecase.type_series, diecase.type_size,
                diecase.wedge.series, diecase.wedge.set_width,
                diecase.typeface_name, json.dumps(diecase.layout)]
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                # Create the table first:
                cursor.execute('CREATE TABLE IF NOT EXISTS matrix_cases ('
                               'diecase_id TEXT UNIQUE PRIMARY KEY, '
                               'type_series TEXT NOT NULL, '
                               'type_size TEXT NOT NULL, '
                               'wedge_series TEXT NOT NULL, '
                               'set_width REAL NOT NULL, '
                               'typeface_name TEXT NOT NULL, '
                               'layout TEXT NOT NULL)')
                # Then add an entry:
                cursor.execute('INSERT INTO matrix_cases ('
                               'diecase_id, type_series, type_size,'
                               'wedge_series, set_width, typeface_name, layout'
                               ') VALUES (?, ?, ?, ?, ?, ?, ?)''', data)
                self.db_connection.commit()
                return True
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def get_diecase(self, diecase_id):
        """Searches for diecase metadata, based on the unique diecase ID."""
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                sql = 'SELECT * FROM matrix_cases WHERE diecase_id = ?'
                cursor.execute(sql, [diecase_id])
                # Return diecase if found:
                diecase = list(cursor.fetchone())
                # De-serialize the diecase layout, convert it back to a list
                raw_layout = json.loads(diecase.pop())
                # Convert the layout
                # Record is now 'character style1,style2... column row units'
                # Make it a list
                layout = []
                for record in raw_layout:
                    # Make a tuple of styles
                    (char, styles, column, row, units) = record
                    # Last field - unit width - was not mandatory
                    # Try to convert it to int
                    # If no unit width is specified, use None instead
                    try:
                        units = int(units)
                    except (IndexError, ValueError):
                        record.append(0)
                    record = (char, tuple(styles), column, int(row), units)
                    layout.append(record)
                    # Record is now:
                    # [character, (style1, style2...), column, row, units]
                # Add the layout back to diecase and return the data
                diecase.append(layout)
                return diecase
            except (TypeError, ValueError, IndexError):
                # No data or cannot process it
                raise exceptions.NoMatchingData
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def delete_diecase(self, diecase):
        """Deletes a diecase with given unique ID from the database."""
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                cursor.execute('DELETE FROM matrix_cases WHERE diecase_id = ?',
                               [diecase.diecase_id])
                return True
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def get_all_ribbons(self):
        """Gets all ribbons stored in database"""
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                cursor.execute('SELECT * FROM ribbons')
                # Check if we got any:
                results = cursor.fetchall()
                if not results:
                    raise exceptions.NoMatchingData
                return results
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def add_ribbon(self, ribbon):
        """Registers a ribbon in our database."""
        data = [ribbon.description, ribbon.customer, ribbon.diecase.diecase_id,
                json.dumps(ribbon.contents)]
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                # Create the table first:
                cursor.execute('CREATE TABLE IF NOT EXISTS ribbons ('
                               'ribbon_id INTEGER PRIMARY KEY AUTOINCREMENT, '
                               'description TEXT, '
                               'customer TEXT, '
                               'diecase_id TEXT NOT NULL, '
                               'contents TEXT NOT NULL)')
                # Then add an entry:
                cursor.execute('INSERT INTO ribbons ('
                               'title, author, customer, diecase_id, contents'
                               ') VALUES (?, ?, ?, ?)''', data)
                self.db_connection.commit()
                return True
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def get_ribbon(self, ribbon_id):
        """Searches for ribbon, based on the unique ribbon ID."""
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                cursor.execute('SELECT * FROM ribbons WHERE ribbon_id = ?',
                               [ribbon_id])
                ribbon = list(cursor.fetchone())
                # Take the last item which is contents
                # De-serialize it, convert it back to a list
                raw_contents = json.loads(ribbon.pop())
                # Add the contents
                ribbon.append(raw_contents)
                # Return the list
                return ribbon
            except (TypeError, ValueError, IndexError):
                # No data or cannot process it
                raise exceptions.NoMatchingData
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def delete_ribbon(self, ribbon):
        """Deletes a ribbon entry from database."""
        # Okay, proceed:
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                cursor.execute('DELETE FROM ribbons WHERE ribbon_id = ?',
                               [ribbon.ribbon_id])
                return True
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def get_all_schemes(self):
        """Gets all font schemes stored in database."""
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                cursor.execute('SELECT * FROM font_schemes')
                # Check if we got any:
                results = cursor.fetchall()
                if not results:
                    raise exceptions.NoMatchingData
                return results
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def add_scheme(self, scheme):
        """Registers a scheme (source text) in our database."""
        data = [scheme.scheme_id, scheme.description, scheme.language,
                json.dumps(scheme.layout)]
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                # Create the table first:
                cursor.execute('CREATE TABLE IF NOT EXISTS font_schemes ('
                               'scheme_id TEXT UNIQUE PRIMARY KEY, '
                               'description TEXT NOT NULL, '
                               'language TEXT NOT NULL, '
                               'scheme_layout TEXT NOT NULL)')
                # Then add an entry:
                cursor.execute('INSERT INTO font_schemes ('
                               'scheme_id, description, language, '
                               'scheme_layout) VALUES (?, ?, ?, ?)''', data)
                self.db_connection.commit()
                return True
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def get_scheme(self, scheme_id):
        """Searches for scheme based on the unique scheme ID."""
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                cursor.execute('SELECT * FROM font_schemes '
                               'WHERE scheme_id = ?', [scheme_id])
                scheme = list(cursor.fetchone())
                layout = scheme.pop()
                scheme.append(json.loads(layout))
                return scheme
            except (TypeError, ValueError, IndexError):
                # No data or cannot process it
                raise exceptions.NoMatchingData
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def delete_scheme(self, scheme):
        """Deletes a scheme with given unique ID from the database."""
        with self.db_connection:
            try:
                cursor = self.db_connection.cursor()
                cursor.execute('DELETE FROM font_schemes WHERE scheme_id = ?',
                               [scheme.scheme_id])
                return True
            except (sqlite3.OperationalError, sqlite3.DatabaseError):
                # Database failed
                raise exceptions.DatabaseQueryError

    def __exit__(self, *args):
        pass



# -*- coding: utf-8 -*-
"""Inventory:

Allows to add, list and delete wedges and diecases.
"""
from . import exceptions as e
from . import matrix_data
from . import wedge_data
from .global_settings import UI


def main_menu():
    """Display the main menu for inventory management"""
    header = ('Matrix case  and wedge manipulation utility for rpi2caster.'
              '\n\nInventory Management menu:\n')
    options = [(e.exit_program, 'Exit',
                'Exits the inventory management'),
               (matrix_data.diecase_operations, 'Matrix operations...',
                'Work on matrix cases'),
               (wedge_data.wedge_operations, 'Wedge operations...',
                'List, add and remove wedge definitions'),
               (wedge_data.generate_wedge_collection,
                'Generate S5 collection',
                'Generates all definitions for S5 wedges for 5...14.75set')]
    while True:
        try:
            UI.menu(options, header=header, footer='')()
        except (e.ReturnToMenu, e.MenuLevelUp):
            pass
        except e.NoMatchingData:
            UI.pause('No matching data found!', UI.MSG_MENU)
        except e.DatabaseQueryError:
            UI.pause('Database query error!', UI.MSG_MENU)
